import pytest

from alphadb import AlphaDB
from alphadb.utils.decorators import conn_test
from alphadb.utils.exceptions import NoConnection


#### Test if a database is connected
class TestDatabaseConnectionDecorator:
    db = AlphaDB()

    #### The function inside the decorator should not execute when no database connection exists. NoConnection exeption should be raised.
    def test_not_connected(self):
        @conn_test
        def conn_test_test(_):
            return True

        with pytest.raises(NoConnection):
            conn_test_test(self.db)

    #### The function inside the decorator should be executed when a database connection exists
    def test_connected(self):
        #### Connect to a database
        self.db.connect(host="localhost", user="root", password="test", database="test")

        @conn_test
        def conn_test_test(_):
            return True

        assert conn_test_test(self.db) == True



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from configparser import ConfigParser, SectionProxy
from cli.utils.exceptions import ConfigIncoplete
from cli.utils.globals import CONFIG_PATH
from typing import overload, Literal

config = ConfigParser()

@overload
def config_get(section: str, key: str) -> str: ...

@overload
def config_get(section: str, key: str, fallback: Literal[True]) -> None: ...

@overload
def config_get(section: str) -> SectionProxy: ...

@overload
def config_get(section:str, *, fallback: Literal[True]) -> None: ...

"Get data from config file"
def config_get(section: str, key: str | None = None, fallback: bool = False) -> SectionProxy | str | None:
    
    config.read(CONFIG_PATH)

    try:
        if key == None:
            return config[section]
        
        return config[section][key]
    except KeyError:
        if fallback: return None
        raise ConfigIncoplete()


"Write to config file"
def config_write(data: dict | list) -> None:

    config.read(CONFIG_PATH)

    #### Loop through root level dict keys
    for key in data:
        config[key] = data[key]

    #### Write changes to config file
    with open(CONFIG_PATH, "w") as configfile:
        config.write(configfile)

    return

"Remove from config file"
def config_remove(section: str, key: str | None = None):

    if key == None:
        config.remove_section(section)
    else:
        config.remove_option(section, key)

    #### Write changes to config file
    with open(CONFIG_PATH, 'w') as configfile:
        config.write(configfile)

    return

"Get all keys of available items in config section"
def config_get_items(key: str) -> list:

    config.read(CONFIG_PATH)

    try:
        return config.items(key)
    except:
        return []



import json
import os
import pickle
import sys

from alphadb import AlphaDB

if __name__ == "__main__":
    #### Id provided by JS script
    instance_id = sys.argv[2]

    db = AlphaDB()

    #### Get cached AlphaDB instance if exitsts
    instance_path = f"./.adbcache/{instance_id}"
    if os.path.isfile(instance_path):
        with open(instance_path, "rb") as f:
            conncreds = pickle.loads(f.read())
            db.connect(host=conncreds["host"], user=conncreds["user"], password=conncreds["password"], database=conncreds["database"], port=conncreds["port"])
    else:
        if not os.path.isdir("./.adbcache"):
            os.mkdir("./.adbcache")  ## To write cache to later

    #### Write AlphaDB instance to cache
    def write_cache(adb_instance):
        with open(instance_path, "wb") as f:
            pickle.dump(adb_instance, f)

    if sys.argv[1] == "check":
        print(json.dumps(db.check()))

    if sys.argv[1] == "connect":
        conncreds = {
            "host": sys.argv[3],
            "user": sys.argv[4],
            "password": sys.argv[5],
            "database": sys.argv[6],
            "port": int(sys.argv[7]),
        }

        print(json.dumps(db.connect(host=conncreds["host"], user=conncreds["user"], password=conncreds["password"], database=conncreds["database"], port=conncreds["port"])))
        write_cache(conncreds)

    if sys.argv[1] == "init":
        print(db.init())

    if sys.argv[1] == "status":
        print(json.dumps(db.status()))

    if sys.argv[1] == "update_queries":
        print(json.dumps(db.update_queries(version_source=json.loads(sys.argv[3]), update_to_version=None if sys.argv[4] == "undefined" else sys.argv[4], no_data=sys.argv[5] == "true")))

    if sys.argv[1] == "update":
        print(db.update(version_source=json.loads(sys.argv[3]), update_to_version=None if sys.argv[4] == "undefined" else sys.argv[4], no_data=sys.argv[5] == "true"))

    if sys.argv[1] == "vacate":
        print(db.vacate(confirm=sys.argv[3] == "true"))

    sys.stdout.flush()



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Literal

from alphadb.utils.common import convert_version_number
from alphadb.utils.types import Method, ValidationIssuesList
from alphadb.verification.compatibility import incompatible_types_with_autoincrement, incompatible_types_with_unique


class VersionSourceVerification:
    def __init__(self, version_source):
        self.version_source = version_source
        self.issues: ValidationIssuesList = []

    def verify(self) -> ValidationIssuesList | Literal[True]:
        """
        Verify the version source.
        Returns True if no issues are identified,
        else it returns a list with the issues represented as strings
        """

        #### Template name
        if not "name" in self.version_source:
            self.issues.append(("CRITICAL", "No rootlevel name was specified"))

        #### Version list
        if not "version" in self.version_source:
            self.issues.append(("LOW", "This version source does not contain any versions"))

        else:
            for i, version in enumerate(self.version_source["version"]):
                if not "_id" in version:
                    version_output = f"Version index {i}"
                    self.issues.append(("CRITICAL", f"{version_output}: Missing a version number"))
                else:
                    try:
                        convert_version_number(version["_id"])
                    except:
                        self.issues.append(("CRITICAL", f"{version['_id']}: Version number is not convertable to an integer"))

                    version_output = f"Version {version['_id']}"
                
                for method in version:
                    match method:
                        case "_id": continue
                        case "createtable": 
                            self.createtable(version["createtable"], version_output)
                        case "altertable":
                            self.altertable(version["altertable"], version_output)
                        case _:
                            self.issues.append(("HIGH", f"{version_output}: Method '{method}' does not exist"))
                    

        return self.issues if not len(self.issues) == 0 else True

    def createtable(self, createtable: dict, version_output: str = "Unknown version"):
        "Verify a single versions createtable"

        if len(createtable) == 0:
            self.issues.append(("LOW", f"{version_output} -> createtable: Does not contain any data"))
        else:
            for table in createtable:
                for column in createtable[table]:
                    #### Primary key
                    if column == "primary_key":
                        if not createtable[table]["primary_key"] in createtable[table]:
                            self.issues.append(("CRITICAL", f"{version_output} -> createtable -> table:{table}: Primary key does not match any column name"))
                        continue

                    #### Columns
                    self.column_compatibility(table, column, createtable[table][column], method="createtable", version_output=version_output)

    def altertable(self, altertable: dict, version_output: str = "Unknown version"):
        "Verify a single versions altertable"

        if len(altertable) == 0:
            self.issues.append(("LOW", f"{version_output} -> altertable: Does not contain any data"))
        else:
            for table in altertable:
                #### Modify column
                if "modifycolumn" in altertable[table]:
                    for column in altertable[table]["modifycolumn"]:
                        self.column_compatibility(table, column, altertable[table]["modifycolumn"][column], method="altertable", version_output=version_output)

                #### Primary key
                # if "primary_key" in table:
                #     if

    def column_compatibility(self, table_name: str, column_name: str, data: dict, method: Method, version_output: str = "Unknown version"):
        "Verify column attribute compatibility"

        #### NULL and AUTO_INCREMENT
        if "null" in data and "a_i" in data:
            self.issues.append(("CRITICAL", f"{version_output} -> {method} -> table:{table_name} -> column:{column_name}: Column attributes NULL and AUTO_INCREMENT are incompatible"))

        #### If type is defined
        if not "type" in data:
            if not "recreate" in data or data["recreate"] == True:
                self.issues.append(("CRITICAL", f"{version_output} -> {method} -> table:{table_name} -> column:{column_name}: Does not contain a column type"))
        else:
            #### Types incompatible with AUTO_INCREMENT
            if data["type"].lower() in incompatible_types_with_autoincrement and "a_i" in data:
                self.issues.append(("CRITICAL", f"{version_output} -> {method} -> table:{table_name} -> column:{column_name}: Column type {data['type']} is incompatible with attribute AUTO_INCREMENT"))

            #### Types incompatible with UNIQUE
            if data["type"].lower() in incompatible_types_with_unique and "unique" in data:
                self.issues.append(("CRITICAL", f"{version_output} -> {method} -> table:{table_name} -> column:{column_name}: Column type {data['type']} is incompatible with attribute UNIQUE"))



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable

from alphadb.utils.decorators import conn_test, init_test
from alphadb.utils.exceptions import DBConfigIncomplete, DBTemplateNoMatch, IncompleteVersionData, MissingVersionData, NeedsConfirmation
from alphadb.utils.globals import CONFIG_TABLE_NAME
from alphadb.utils.query.default_data import create_default_data
from alphadb.utils.query.table.altertable import altertable
from alphadb.utils.query.table.createtable import createtable

from mysql.connector import MySQLConnection
from mysql.connector.cursor import MySQLCursor


class AlphaDB:
    db_name: str
    connection: MySQLConnection | None = None
    cursor: Callable[..., MySQLCursor]

    def __init__(self):
        self.connection = None

    @conn_test
    def check(self):
        with self.cursor() as cursor:
            current_version = None

            #### Check if the config table (adb_conf) exists
            cursor.execute(
                f"SELECT table_name FROM information_schema.tables WHERE table_schema = %s AND table_name = %s",
                (self.db_name, CONFIG_TABLE_NAME),
            )

            table_check = cursor.fetchall()

            fetched = None
            if table_check:
                cursor.execute(
                    f"SELECT version FROM {CONFIG_TABLE_NAME} WHERE db = %s",
                    (self.db_name,),
                )
                fetched = cursor.fetchone()
                if fetched:
                    current_version = fetched[0]

            #### Check True means the database is ready for use
            check = True if table_check and fetched else False

        return {"check": check, "current_version": current_version}

    def connect(self, host: str, password: str, user: str, database: str, port: int = 3306) -> bool:
        self.connection = MySQLConnection(
            host=host,
            user=user,
            password=password,
            port=port,
            database=database,
            buffered=True,
        )
        self.connection.autocommit = True
        self.cursor = self.connection.cursor
        self.db_name = database

        return self.check()["check"] == True

    @conn_test
    def init(self):
        #### Check if the database is already initialized
        if self.check()["check"] == True:
            return "already-initialized"

        try:
            with self.cursor() as cursor:
                #### Create configuration table
                cursor.execute(f"CREATE TABLE IF NOT EXISTS {CONFIG_TABLE_NAME} (db VARCHAR(100) NOT NULL, version VARCHAR(50) NOT NULL, template VARCHAR(50) NULL, PRIMARY KEY (db)) ENGINE = InnoDB;")

                cursor.execute(
                    f"INSERT INTO {CONFIG_TABLE_NAME} (db, version) values (%s, %s)",
                    (self.db_name, "0.0.0"),
                )
        except Exception as e:
            raise Exception(e)

        return True

    @conn_test
    def status(self):
        current_version = None
        template = None

        with self.cursor() as cursor:
            #### Check if the config table (adb_conf) exists
            cursor.execute(
                f"SELECT table_name FROM information_schema.tables WHERE table_schema = %s AND table_name = %s",
                (self.db_name, CONFIG_TABLE_NAME),
            )
            table_check = cursor.fetchall()

            #### If it exists, get current version
            fetched = None
            if table_check:
                cursor.execute(
                    f"SELECT version, template FROM {CONFIG_TABLE_NAME} WHERE db = %s",
                    (self.db_name,),
                )
                fetched = cursor.fetchone()
                if fetched:
                    current_version = fetched[0]
                    template = fetched[1]

            #### Check True means the database is initialized
            check = True if table_check and fetched else False

        return {
            "init": check,
            "version": current_version,
            "name": self.db_name,
            "template": template,
        }

    @conn_test
    @init_test
    def update_queries(self, version_source, update_to_version=None, no_data=False):
        print(version_source)

        #### Some error handling
        if version_source == None:
            raise MissingVersionData()
        else:
            if not "version" in version_source or not "name" in version_source:
                raise IncompleteVersionData()

        #### Start update process
        database_version = None
        with self.cursor() as cursor:
            try:
                cursor.execute(
                    f"SELECT version, template FROM {CONFIG_TABLE_NAME} WHERE db = %s",
                    (self.db_name,),
                )
                db_data = cursor.fetchone()

                if not db_data == None:
                    #### Check if the database template matches
                    if not version_source["name"] == db_data[1]:
                        #### If no template is defined, use the current one
                        if db_data[1] == None:
                            cursor.execute(
                                f"UPDATE {CONFIG_TABLE_NAME} SET template = '{version_source['name']}' WHERE db = %s",
                                (self.db_name,),
                            )
                        else:
                            raise DBTemplateNoMatch()
                    database_version = db_data[0]  ## Get database version
            except Exception as e:
                raise e

            #### If no database version is returned from the database, config might be broken
            if database_version == None or database_version == "":
                raise DBConfigIncomplete(missing="version")

            #### Get the latest database version
            latest = max(version_source["version"], key=lambda x: int(x["_id"].replace(".", "")))["_id"]
            database_version_latest = latest if update_to_version == None else update_to_version

            ### Check if database needs to be updated
            try:
                if not int(database_version_latest.replace(".", "")) > int(database_version.replace(".", "")):
                    return "up-to-date"

            #### If db version number can not be formatted to int, it's invalid
            except ValueError:
                raise DBConfigIncomplete(missing="version")

            try:
                #### List to be populated with queries
                query_list = []

                #### Loop through update data
                for version in version_source["version"]:
                    #### Check if version number is larger than current version
                    if int(version["_id"].replace(".", "")) <= int(database_version.replace(".", "")):
                        continue

                    #### Continue if latest version is current
                    if int(database_version_latest.replace(".", "")) < int(version["_id"].replace(".", "")):
                        continue

                    #### Create tables
                    if "createtable" in version:
                        for table in version["createtable"]:
                            query = createtable(version_source=version_source, table_name=table, version=version["_id"])
                            query_list.append(query)

                    #### Alter tables
                    if "altertable" in version:
                        for table in version["altertable"]:
                            query = altertable(version_source=version_source, table_name=table, version=version["_id"])
                            query_list.append(query)

                    #### Insert default data
                    if no_data == False:
                        if "default_data" in version:
                            for table in version["default_data"]:
                                for item in version["default_data"][table]:
                                    query = create_default_data(table_name=table, item=item)
                                    query_list.append(query)
            except KeyError as e:
                raise e
            except Exception as e:
                raise e

            query_list.append(
                (
                    f"UPDATE `{CONFIG_TABLE_NAME}` SET version=%s WHERE `db` = %s",
                    (database_version_latest, self.db_name),
                )
            )

            return query_list

    @conn_test
    @init_test
    def update(self, version_source, update_to_version=None, no_data=False):
        queries = self.update_queries(version_source=version_source, update_to_version=update_to_version, no_data=no_data)

        if queries == "up-to-date":
            return "up-to-date"

        with self.cursor() as cursor:
            for query in queries:
                #### If type is typle, data is passed with the query
                if type(query) == tuple:
                    cursor.execute(*query)
                else:
                    cursor.execute(query)

        return True

    @conn_test
    def vacate(self, confirm=False):
        if not confirm == True:
            raise NeedsConfirmation()

        with self.cursor() as cursor:
            cursor.execute("SET foreign_key_checks = 0;")
            cursor.execute("show tables;")
            tables = cursor.fetchall()

            for t in tables:
                cursor.execute(f"DROP TABLE {t[0]}")

            cursor.execute("SET foreign_key_checks = 1;")

        return True

    @conn_test
    def export(self):
        with self.cursor() as cursor:
            cursor.execute("show tables;")
            tables = cursor.fetchall()

            data = {}

            for t in tables:
                cursor.execute(f"SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = N'{t[0]}'")
                data[t[0]] = {"columns": [t[3] for t in cursor.fetchall()]}

                cursor.execute(f"SELECT * FROM {t[0]}")
                data[t[0]]["data"] = cursor.fetchall()

        return data



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cli.utils.common import print_title
from cli.utils.connect import get_mysql_creds
from alphadb import AlphaDB, VersionSourceVerification
from alphadb.utils.exceptions import DBTemplateNoMatch, IncompleteVersionData, MissingVersionData, DBConfigIncomplete
from mysql.connector import DatabaseError, InterfaceError
from cryptography.fernet import Fernet
from cli.utils.config import config_get, config_write 
from cli.utils.common import console
from cli.utils.decorators import connection_check
from cli.utils.version_source import select_version_source, vs_path_to_json
from inquirer import prompt, Confirm
from cli.utils import globals

def connect():
    "Connect to a database"

    print_title("connect")
        
    globals.db = AlphaDB()
    creds = get_mysql_creds()

    try:
        globals.db.connect(
            host=creds["host"],
            user=creds["user"],
            password=creds["password"],
            database=creds["database"],
            port=creds["port"],
        )

    except (DatabaseError, InterfaceError) as e:
        if hasattr(e, "msg"): console.print(f"\n[red]{e.msg}[/red]\n")
        else: console.print(f"\n[red]{e}[/red]\n")
        return

    pass_bytes = creds["password"].encode()
    f = Fernet(config_get("CONFIG", "secret"))
    pass_encrypted = f.encrypt(pass_bytes)

    config_write({
        "DB_SESSION": {
            "host": creds["host"],
            "user": creds["user"],
            "password": pass_encrypted,
            "database": creds["database"],
            "port": creds["port"],
        }
    })

    console.print(f'\n[green]Successfully connected to database:[/green] [cyan]"{creds["database"]}"[/cyan]\n')


    return

@connection_check()
def init():
    "Initialize database"
    
    #### Initialize loader
    with console.status("[cyan]Getting the database ready[/cyan]", spinner="bouncingBall") as _:
        init = globals.db.init()

    if init == "already-initialized":
        console.print("[yellow]The database is already initialized[/yellow]\n")
        return

    if init == True:
        console.print("[green]Database successfully initialized[/green]\n")
        return

@connection_check()
def status():
    "Get database status"
    print_title("database status")

    check = globals.db.status()

    print(f'Database: {check["name"]}')
    print(f'Template: {check["template"]}')
    if check["init"] == True:
        console.print("Status: ", "[cyan]Initialized[/cyan]")
    else:
        console.print("Status: ", "[yellow]Uninitialized[/yellow]")
    print(f'Version: {check["version"]}\n')

    return

@connection_check()
def update(nodata=False):
    "Update database"

    print_title("update")

    #### Initialize loader
    with console.status("[cyan]Checking database[/cyan]", spinner="bouncingBall") as loader:

        #### Check database status
        status = globals.db.status()
        if status["init"] == False:
            console.print(f"[yellow]Database [cyan]{status['name']}[/cyan] has not yet been initialized[/yellow]\n")
            return

    try:

        version_source_path = select_version_source()
    
        #### If source path is None, user probably aborted
        if version_source_path == None: return
        
        with console.status("[cyan]Reading version source[/cyan]", spinner="bouncingBall") as loader:
           
            version_information = vs_path_to_json(version_source_path)

            #### If version information is None, an error has likely occured and we should not proceed
            if version_information == None: return

            loader.update("[cyan]Running updates on the database[/cyan]")

            update = globals.db.update(version_source=version_information, no_data=nodata)

            if update:
                if update == "up-to-date":
                    console.print(f"[blue]Database is already the latest version [cyan]({status['version']})[/cyan][/blue]\n")
                    return

                console.print("[green]Database successfully updated to the latest version[/green]\n")
            else:
                console.print("[red]An error occured[/red]\n")

    except (DBTemplateNoMatch, IncompleteVersionData, MissingVersionData, DBConfigIncomplete) as e:
        console.print(f"[red]{e}[/red]\n")

@connection_check()
def vacate(confirm=False):
    "Empty database"
    print_title("vacate")
    
    #### Only works when confirm==True is specified
    if not confirm == True:
        console.print(f"[yellow]The vacate function requires the[/yellow] [red]--confirm[/red] [yellow]option.[/yellow]")
        print("This is a safety feature that hopefully prevents unintended data loss.\n\nDon't worry, you'll still be prompted for confirmation!\n")
        return

    console.print(f"[yellow]The vacate function[/yellow] [red]deletes all data in the database[/red].")
    console.print("[yellow]This action can [red]NOT[/red] be undone.[/yellow]\n")
    
    #### Prompt user for confirmation
    questions = [Confirm("confirm", message="Are you absolutely sure you want to completely delete all data?")]
    answers = prompt(questions)

    #### If answers is None, user likely aborted
    if answers == None: return

    if answers["confirm"]:
        with console.status("[cyan]Removing all data from database[/cyan]", spinner="bouncingBall") as _:

            #### Empty db
            if globals.db.vacate(confirm=confirm):
                console.print("[green]The database had successfully been emptied.[/green]\n")
    else:
        console.print("[cyan]Not empying[/cyan]\n")
    return

def verify_version_source():
    print_title("verify version source")
    
    version_source_path = select_version_source()

    #### If version source path is None, user likely aborted
    if version_source_path == None: return

    version_source = vs_path_to_json(version_source_path)

    #### If version source is None, an error likely occured and we should not proceed
    if version_source == None: return

    verification = VersionSourceVerification(version_source)

    output = verification.verify()

    if output == True:
        console.print(f"[green]Version source at [blue]{version_source_path}[/blue] verified without errors[/green]\n")
        return

    console.print(f"Version source at [blue]{version_source_path}[/blue] has [red]{len(output)} errors[/red]\n\n")

    for issue in sorted(output):
            

        if ":" in issue[1]:
            issue_path, issue_text = issue[1].rsplit(': ', 1)
        else:
            issue_path = ""
            issue_text = issue[1]

        if issue[0] == "LOW":
            console.print("[black on white]LOW VULNERABILITY: [/black on white]", f"[cyan]{issue_path}[/cyan]", issue_text)

        if issue[0] == "HIGH":
            console.print("[white on yellow]HIGH VULNERABILITY: [/white on yellow]", f"[cyan]{issue_path}[/cyan]", f"[yellow]{issue_text}[/yellow]")

        if issue[0] == "CRITICAL":
            console.print("[white on red]CRITICAL: [/white on red]", f"[cyan]{issue_path}[/cyan]", f"[red]{issue_text}[/red]")
        
    console.line()



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, get_args

from alphadb.utils.exceptions import IncompatibleColumnAttributes, IncompleteVersionObject
from alphadb.utils.types import DatabaseColumnType
from alphadb.verification.compatibility import incompatible_types_with_unique, incompatible_types_with_autoincrement

def definecolumn(column_name: str, column_type: DatabaseColumnType, null: bool = False, length: Optional[int] = None, unique: bool = False, default: Optional[str | int] = None, auto_increment: bool = False):
    if not column_type.upper() in get_args(DatabaseColumnType):
        raise ValueError(f"Column type {column_type} is not (yet) supported.")

    query = f" {column_name} {column_type}"

    if not length == None:
        query += f"({length})"

    query += " NULL" if null else " NOT NULL"

    if unique:
        query += " UNIQUE"

    if default:
        query += f" DEFAULT '{default}'"

    #### Auto increment
    if auto_increment:
        query += " AUTO_INCREMENT"

    return query


def prepare_definecolumn_data(table_name: str, column: str, table_data: dict, version: str):
    #### If iteration is not of type Dict, it is not a column and should be handled later
    if not isinstance(table_data[column], dict) or column == "foreign_key":  ## Foreign key IS an object, but has to be handled later
        return None

    #### A column type must be defined
    if not "type" in table_data[column]:
        raise IncompleteVersionObject(key="type", object=f"Version {version}->{table_name}->{column}")

    qlength = table_data[column]["length"] if "length" in table_data[column] else None
    qnull = table_data[column]["null"] if "null" in table_data[column] else False
    qunique = table_data[column]["unique"] if "unique" in table_data[column] else False
    qdefault = table_data[column]["default"] if "default" in table_data[column] else None
    qautoincrement = table_data[column]["a_i"] if "a_i" in table_data[column] else False

    #### Check for column type compatibility with AUTO_INCREMENT
    if table_data[column]["type"].lower() in incompatible_types_with_autoincrement and qautoincrement == True:
        raise IncompatibleColumnAttributes(f"type=={table_data[column]['type']}", "AUTO_INCREMENT", version=f"Version {version}->{table_name}->{column}")

    #### Check for column type compatibility with UNIQUE
    if table_data[column]["type"].lower() in incompatible_types_with_unique and qunique == True:
        raise IncompatibleColumnAttributes(f"type=={table_data[column]['type']}", "UNIQUE", version=f"Version {version}->{table_name}->{column}")

    #### Null will be ignored by the database engine when AUTO_INCREMENT is specified
    if qnull == True and qautoincrement == True:
        raise IncompatibleColumnAttributes("NULL", "AUTO_INCREMENT", version=f"Version {version}->{table_name}->{column}")

    return {"length": qlength, "null": qnull, "unique": qunique, "default": qdefault, "auto_increment": qautoincrement}



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json


def create_default_data(table_name: str, item: dict):
    #### Definitions for query generation
    keys = ""
    values = ""

    #### Create keys and values strings
    for c, key in enumerate(item):
        #### If value is None, skip
        if item[key] == None:
            continue

        keys += key

        #### Check if value is JSON data
        if type(item[key]) is dict:
            values += f"'{json.dumps(item[key])}'"
        elif type(item[key]) is int:
            values += f"{item[key]}"
        elif type(item[key]) is bool:
            values += "true" if item[key] == True else "false"
        else:
            values += f"'{item[key]}'"

        #### Add a comma (trailing comma will be removed after loop, less complex logic)
        keys += ","
        values += ","

    #### Remove trailing comma
    keys = keys[:-1]
    values = values[:-1]

    #### Return query
    return f"INSERT INTO `{table_name}` ({keys}) VALUES ({values});"



# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional
import typer
from cli import __app_name__, __version__, commands

app = typer.Typer()

@app.command(help="Initialize the currently active database")
def init() -> None:
    commands.init()

@app.command(help="Show to status of the currently active database")
def status() -> None:
    commands.status()

@app.command(help="Update the database (requires a version source)")
def update(
    nodata: Optional[bool] = typer.Option(
        False,
        "--no-data",
        help="Update only to the database structure, but do not include default data",
    )
) -> None:
    commands.update(nodata=nodata if not nodata == None else False)

@app.command(help="Irriversibally deletes ALL data in the database")
def vacate(
    confirm: Optional[bool] = typer.Option(
        None,
        "--confirm",
        help="Needs to be specified. This is a safety feature. Only with this option specified the vacate function will be called",
    )
) -> None:
    commands.vacate(confirm=confirm if not confirm == None else False)

@app.command(help="Connect to a new database")
def connect() -> None:
    commands.connect()

@app.command("verify", help="Verify version source")
def verify_version_source() -> None:
    commands.verify_version_source()

def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__} {__version__}")
        raise typer.Exit()
    return

@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-V",
        help="Show the application's version and exit.",
        callback=version_callback,
        is_eager=True,
    )
) -> None:
    return



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cli import __app_name__, cli as cli_setup
from cli.utils.decorators import config_check
from cli.utils.common import console, clear

DEV = True

def raise_error(e):
    if DEV: raise e
    else: 
        if hasattr(e, "msg"):
            console.print(e.msg, style="white on red")
        else:
            console.print(f"[white]{e}[/white]", style="white on red")

@config_check
def main():
    clear()
    try:
        cli_setup.app(prog_name=__app_name__)
    except Exception as e:
        raise_error(e)

if __name__ == "__main__":
    main()



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from configparser import ConfigParser
from typing import Callable

from alphadb import AlphaDB
from cryptography.fernet import Fernet
from simpleUID import urlsafe

from cli.utils import globals
from cli.utils.common import console
from cli.utils.config import config_get


def config_check(func: Callable) -> Callable:
    "Initialize config when it's not yet initialized"

    def _(*args, **kwargs):
        config = ConfigParser()
        if os.path.isfile(globals.CONFIG_PATH):
            config.read(globals.CONFIG_PATH)

        if not "CONFIG" in config:
            config["CONFIG"] = {}
        if not "secret" in config["CONFIG"]:
            config["CONFIG"]["secret"] = urlsafe("padding")

        with open(globals.CONFIG_PATH, "w+") as configfile:
            config.write(configfile)

        return func(*args, **kwargs)

    return _


def connection_check() -> Callable:
    "Check if a database connection exists"

    def _(func: Callable) -> Callable:
        def _wrapper(*args, **kwargs):
            if globals.db.connection == None:
                conn = config_get("DB_SESSION", fallback=True)
                main_config = config_get("CONFIG")
                secret = main_config["secret"] if "secret" in main_config else None

                if conn and secret:
                    globals.db = AlphaDB()

                    try:
                        globals.db.connect(
                            host=conn["host"],
                            user=conn["user"],
                            password=Fernet(secret)
                            .decrypt(conn["password"][1:-1])
                            .decode(),  ## Decrypt
                            database=conn["database"],
                            port=conn["port"],
                        )

                        globals.db.connection.autocommit = True
                    except:
                        console.print(
                            "[red]Unable to authorize using saved credentials. Please reconnect to the database.[/red]"
                        )
                        return

                    return func(*args, **kwargs)

                console.print("\n[yellow]No database connection found.[/yellow]\n")

        return _wrapper

    return _



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="alphadb",

    #### ! Do not change the version number manually, this will be handled by the build script
    version="v1.0.0b6",

    author="Wibo Kuipers",
    author_email="wibokuip@gmail.com",
    description="MySQL Database versioning toolset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/w-kuipers/alphadb",
    project_urls={
        "Bug Tracker": "https://github.com/w-kuipers/alphadb/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=["mysql-connector-python==8.2.0"],
)
